# Extremely simplified type stubs for git and git.lfs.
# Extend as needed.
#
# Copyright (C) 2021 Collabora Ltd
# Copyright (C) 2021 Andrej Shadura <andrew.shadura@collabora.co.uk>
#
# SPDX-License-Identifier: GPL-2.0-or-later

class git:
    class lfs:
        def __init__(*args): ...
        def __call__(self, *args, **kwargs): ...
        def smudge(self, *args, **kwargs): ...
        def clean(self, *args, **kwargs): ...
    def __init__(*args, **kwargs): ...
    def __call__(self, *args, **kwargs): ...
